<?php

declare(strict_types=1);

namespace Drupal\tailwindcss_utility\Form;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\layout_builder\LayoutTempstoreRepositoryInterface;
use Drupal\layout_builder\Section;
use Drupal\layout_builder\SectionComponent;
use Drupal\layout_builder\SectionStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

final class LayoutFormsAlter implements ContainerInjectionInterface {

  use DependencySerializationTrait;
  use StringTranslationTrait;

  /**
   * The layout tempstore repository.
   *
   * @see https://www.drupal.org/project/drupal/issues/3110266
   * @phpstan-ignore-next-line
   */
  protected LayoutTempstoreRepositoryInterface $layoutTempstoreRepository;

  public function __construct(LayoutTempstoreRepositoryInterface $layout_tempstore_repository) {
    $this->layoutTempstoreRepository = $layout_tempstore_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('layout_builder.tempstore_repository')
    );
  }

  // Common methods.
  private function getSection(FormStateInterface $form_state): ?Section {
    $build_info = $form_state->getBuildInfo();
    [$section_storage, $delta] = $build_info['args'];

    $sections = $section_storage->getSections();
    return \array_key_exists($delta, $sections) ? $sections[$delta] : NULL;
  }

  private function getLayoutStorage(FormStateInterface $form_state): SectionStorageInterface {
    $build_info = $form_state->getBuildInfo();
    return $build_info['args'][0];
  }

  /**
   * Get form element for Tailwind classes.
   *
   * @return mixed[]
   */
  private function getClassesElement($default_value): array {
    return [
      '#type' => 'textfield',
      '#title' => $this->t('Tailwind classes'),
      '#description' => $this->t('Enter class names separated with a space. For more information see the @link.', [
        '@link' => Link::fromTextAndUrl($this->t('Tailwind docs'), Url::fromUri('https://tailwindcss.com/docs'))->toString(),
      ]),
      '#autocomplete_route_name' => 'tailwindcss_utility.class_autocomplete',
      '#autocomplete_route_parameters' => [
        'include_core' => TRUE,
      ],
      '#default_value' => $default_value,
    ];
  }

  // Section form methods.
  public function alterSectionForm(array &$form, FormStateInterface $form_state): void {
    $classes = '';
    $section = $this->getSection($form_state);
    if ($section instanceof Section) {
      $settings = $section->getLayoutSettings();
      if (\array_key_exists('tailwind_classes', $settings)) {
        $classes = $settings['tailwind_classes'];
      }
    }

    $form['layout_settings']['tailwind_classes'] = $this->getClassesElement($classes);
    $form['#submit'][] = [$this, 'submitSectionForm'];
  }

  public function submitSectionForm($form, FormStateInterface $form_state): void {
    $section = $this->getSection($form_state);
    $settings = $section->getLayoutSettings();

    $tailwind_classes = $form_state->getValue([
      'layout_settings',
      'tailwind_classes',
    ]);
    if (\strlen($tailwind_classes) === 0) {
      unset($settings['tailwind_classes']);
    }
    else {
      $settings['tailwind_classes'] = $tailwind_classes;
    }

    $section->setLayoutSettings($settings);
    $this->layoutTempstoreRepository->set($this->getLayoutStorage($form_state));
  }

  // Block form methods.
  private function getComponent(FormStateInterface $form_state): SectionComponent {
    $section = $this->getSection($form_state);
    $component = $form_state->get('layout_builder__component');
    if (\is_null($component)) {
      $uuid = $form_state->getBuildInfo()['args'][3];
      $component = $section->getComponent($uuid);
    }
    return $component;
  }

  public function alterBlockForm(array &$form, FormStateInterface $form_state): void {
    $configuration = $this->getComponent($form_state)->get('configuration');
    if (!\array_key_exists('settings', $configuration)) {
      $configuration['settings'] = [];
    }
    $tailwind_classes = \array_key_exists('tailwind_classes', $configuration['settings']) ? $configuration['settings']['tailwind_classes'] : NULL;

    $form['settings']['tailwind_classes'] = $this->getClassesElement($tailwind_classes);
    $form['#submit'][] = [$this, 'submitBlockForm'];
  }

  public function submitBlockForm($form, FormStateInterface $form_state): void {
    $component = $this->getComponent($form_state);
    $configuration = $component->get('configuration');
    if (!\array_key_exists('settings', $configuration)) {
      $configuration['settings'] = [];
    }

    $tailwind_classes = $form_state->getValue(['settings', 'tailwind_classes']);
    if (\strlen($tailwind_classes) === 0) {
      unset($configuration['settings']['tailwind_classes']);
    }
    else {
      $configuration['settings']['tailwind_classes'] = $tailwind_classes;
    }
    $component->setConfiguration($configuration);

    $this->layoutTempstoreRepository->set($this->getLayoutStorage($form_state));
  }

}
