<?php

declare(strict_types=1);

namespace Drupal\tailwindcss_utility;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Query\Merge;
use Drupal\Core\Site\Settings;
use Psr\Log\LoggerInterface;

final class RuleStorage {

  private const DEFAULT_CACHE_KEY = '_default';

  /**
   * The module config.
   */
  private Config $config;

  /**
   * The cache bin.
   */
  private CacheBackendInterface $cache;

  /**
   * The tailwindcss_utility logger channel.
   */
  private LoggerInterface $logger;

  /**
   * Creates a new RuleStorage object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    CacheBackendInterface $cache,
    LoggerInterface $logger
  ) {
    $this->config = $config_factory->getEditable('tailwindcss_utility.config');
    $this->cache = $cache;
    $this->logger = $logger;
  }

  /**
   * Gets default rules from file storage.
   *
   * @return mixed[]
   */
  private function getDefaultRules(): array {
    $cached = $this->cache->get(self::DEFAULT_CACHE_KEY);
    if (!empty($cached->data)) {
      return $cached->data;
    }

    $json_file = \DRUPAL_ROOT . '/' . Settings::get('tailwind_css_json_path');
    if (!\file_exists($json_file) || !\is_file($json_file)) {
      $this->logger->error('JSON file with tailwind styles not found at the given location: %path.', [
        '%path' => $json_file,
      ]);
      return [];
    }

    $output = [];

    $contents = \file_get_contents($json_file);
    $decoded = \json_decode($contents, TRUE);

    if (\is_null($decoded) || !\is_array($decoded)) {
      $this->logger->error('Invalid JSON file with tailwind styles at the given location: %path.', [
        '%path' => $json_file,
      ]);
      return [];
    }

    foreach ($decoded as $selector => $rules) {
      // We want single classes only.
      if (!\str_starts_with($selector, '.')) {
        continue;
      }
      $class = \stripslashes(\substr($selector, 1));
      $output[$class] = $rules;
    }

    $this->cache->set(self::DEFAULT_CACHE_KEY, $output);
    return $output;
  }

  /**
   * Get all custom rules.
   *
   * @return mixed[]
   */
  private function getCustomRules(): array {
    $rules = $this->config->get('custom_rules');
    return \is_array($rules) ? $rules : [];
  }

  /**
   * Rule getter.
   *
   * @param string[] $selectors
   *
   * @return mixed[]
   */
  public function getRules(array &$classes): array {
    // Load all rules first and filter by classes to maintain Tailwind order.
    $rules = $this->getDefaultRules();

    // Find custom rules, add in place of existing to apply possible overrides.
    foreach ($this->getCustomRules() as $class => $custom_rules) {
      $rules[$class] = $custom_rules;
    }

    // Flip classes first as array_key_exists is much faster than in_array
    // or array_search even with flip if searching many times afterwards.
    $flipped_classes = \array_flip($classes);
    foreach (\array_keys($rules) as $class) {
      if (!\array_key_exists($class, $flipped_classes)) {
        unset($rules[$class]);
        continue;
      }
      unset($classes[$flipped_classes[$class]]);
    }

    return $rules;
  }

  /**
   * Get rules for a single class.
   *
   * @return string[]
   */
  public function getSingleClassRules(string $class): ?array {
    // @todo This could be optimized in case of a single rule
    // if ever needed, this method is run only from admin settings.
    $classes = [$class];
    $rules = $this->getRules($classes);
    return \array_key_exists($class, $rules) ? $rules[$class] : NULL;
  }

  /**
   * Search for classes.
   *
   * @return string[]
   */
  public function classSearch(string $string, bool $include_core = FALSE): array {
    $results = \array_keys($this->getCustomRules());
    $results = \array_filter($results,
      static fn ($class) => \str_starts_with($class, $string)
    );
    \sort($results);

    if ($include_core) {
      $default_classes = \array_keys($this->getDefaultRules());
      $default_classes = \array_filter($default_classes,
        static fn ($class) => \str_starts_with($class, $string)
      );
      $results = \array_unique(\array_merge($results, $default_classes));
      \sort($results);
    }

    // Return 10 results.
    if (\count($results) > 10) {
      $results = \array_slice($results, 0, 10);
    }

    return $results;
  }

  /**
   * Rule setter.
   *
   * @see Drupal\Core\Database\Query\Merge::execute().
   */
  public function addRule(string $class, array $rule): int {
    $rules = $this->getCustomRules();
    if (\array_key_exists($class, $rules)) {
      $return = Merge::STATUS_UPDATE;
    }
    else {
      $return = Merge::STATUS_INSERT;
    }
    $rules[$class] = $rule;
    $this->config->set('custom_rules', $rules);
    $this->config->save();
    return $return;
  }

  /**
   * Delete rule.
   *
   * @return bool
   *   TRUE on success, FALSE otherwise.
   */
  public function deleteRule(string $class): bool {
    $rules = $this->getCustomRules();
    if (\array_key_exists($class, $rules)) {
      unset($rules[$class]);
      $this->config->set('custom_rules', $rules);
      $this->config->save();
      return TRUE;
    }

    return FALSE;
  }

}
