<?php

declare(strict_types=1);

namespace Drupal\tailwindcss_utility\StackMiddleware;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Render\HtmlResponse;
use Drupal\tailwindcss_utility\RuleStorage;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Executes the page caching before the main kernel takes over the request.
 *
 * @phpstan-ignore-next-line
 *
 * This class should be extendable.
 */
class Tailwind implements HttpKernelInterface {

  /**
   * The wrapped HTTP kernel.
   *
   */
  protected HttpKernelInterface $httpKernel;

  /**
   * The cache bin.
   *
   */
  protected CacheBackendInterface $cache;

  /**
   * The tailwindcss_utility logger channel.
   */
  protected LoggerInterface $logger;

  /**
   * Tailwind rule storage service.
   */
  protected RuleStorage $ruleStorage;

  /**
   * The prefixAllowList to find tailwind classes.
   *
   */
  protected string $prefixAllowList = <<<'EOF'
absolute
accent
align
animate
antialiased
appearance
aspect
auto
backdrop
basis
bg
block
blur
border
bottom
box
break
brightness
capitalize
caret
clear
col
columns
content
contents
contrast
cursor
decoration
delay
diagonal
divide
drop
duration
ease
fill
filter
fixed
flex
float
flow
font
from
gap
grayscale
grid
grow
h
hidden
hue
indent
inline
inset
invert
invisible
isolate
isolation
italic
items
justify
leading
left
line
lining
list
lowercase
m
max
mb
min
mix
ml
mr
mt
mx
my
no
normal
not
object
oldstyle
opacity
order
ordinal
origin
outline
overflow
overline
overscroll
p
pb
pl
place
placeholder
pointer
pr
proportional
pt
px
py
relative
resize
right
ring
rotate
rounded
row
saturate
scale
scroll
select
self
sepia
shadow
shrink
skew
slashed
snap
space
sr
stacked
static
sticky
stroke
subpixel
table
tabular
text
to
top
touch
tracking
transform
transition
translate
truncate
underline
uppercase
via
visible
w
whitespace
will
z
EOF;

  /**
   * Constructs a PageCache object.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   The decorated kernel.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache bin.
   * @param \Drupal\tailwind\RuleStorage $rule_storage
   *   Tailwind rule storage service.
   */
  public function __construct(
    HttpKernelInterface $http_kernel,
    CacheBackendInterface $cache,
    LoggerInterface $logger,
    RuleStorage $rule_storage
  ) {
    $this->httpKernel = $http_kernel;
    $this->cache = $cache;
    $this->logger = $logger;
    $this->ruleStorage = $rule_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function handle(Request $request, $type = self::MASTER_REQUEST, $catch = TRUE) {
    $response = $this->httpKernel->handle($request, $type, $catch);

    // Only handle HTML responses for now.
    // @todo What about AjaxResponse?
    if (!($response instanceof HtmlResponse)) {
      return $response;
    }

    $content = $response->getContent();
    if (\strpos($content, 'app--tailwind-utilities.css?tailwind-dynamic=1') === FALSE) {
      return $response;
    }

    // Find all classes, classes used in Alpine are out of scope and should be
    // added via data-class="".
    $pattern = '/ (class|data-class)="([^"]+)"/';
    \preg_match_all($pattern, $content, $matches, \PREG_SET_ORDER);

    if (empty($matches)) {
      return $response;
    }

    $classes = [];
    foreach ($matches as $match) {
      $single_classes = \explode(' ', $match[2]);
      foreach ($single_classes as $class) {
        if (!empty($class)) {
          $classes[$class] = TRUE;
        }
      }
    }
    \ksort($classes);

    // Add pseudo-class for things we want to always keep.
    $classes['keep:always'] = TRUE;

    $prefixes = \array_flip(\explode("\n", $this->prefixAllowList));

    // Filter classes down to a smaller set.
    $filtered_classes = [];
    foreach ($classes as $class => $status) {
      // Ignore toolbar.
      if (\strpos($class, 'toolbar') === 0) {
        continue;
      }

      // Including : or starting with - is always Tailwind.
      if (\strpos($class, ':') !== FALSE || \strpos($class, '-') === 0) {
        $filtered_classes[] = $class;
        continue;
      }

      // Ensure prefix is one of the tailwind ones.
      [$prefix] = \explode('-', $class, 2);
      // @phpstan-ignore-next-line
      if (isset($prefixes[$prefix])) {
        $filtered_classes[] = $class;
        continue;
      }
    }

    if (\count($filtered_classes) === 0) {
      return $response;
    }

    $tailwind_styles = $this->getStyleRules($filtered_classes);
    if (\count($tailwind_styles) === 0) {
      return $response;
    }

    // Process style rules.
    $inline_styles = [];
    foreach ($tailwind_styles as $style_items) {
      foreach ($style_items as $style_item) {
        $media = \array_key_exists('atrule', $style_item) ? $style_item['atrule'] : '_ALL';
        $inline_styles[$media][] = $style_item['rule'];
      }
    }

    $inline_styles_str = '';
    foreach ($inline_styles as $atrule => $rules) {
      if ($atrule == '_ALL') {
        $inline_styles_str .= \implode("\n", $rules);
        $inline_styles_str .= \PHP_EOL;
        continue;
      }

      [$prefix, $suffix] = \explode("\n", $atrule, 2);
      $inline_styles_str .= $prefix . \PHP_EOL . "  ";
      $inline_styles_str .= \implode("\n  ", $rules);
      $inline_styles_str .= $suffix . \PHP_EOL;
      $inline_styles_str .= \PHP_EOL;
    }

    $styles = '<style>' . $inline_styles_str . '</style>';
    $content = \preg_replace('|<link rel="stylesheet" media="all" href=".*/app--tailwind-utilities.css\?tailwind-dynamic=1.*" />|i', $styles, $content);
    $response->setContent($content);
    return $response;
  }

  /**
   * Returns the tailwind style rules needed for the given classes.
   *
   * @param string[] $classes
   *   The classes to return the tailwind styles for.
   * @return array
   *   An array of arrays of style rule objects.
   */
  protected function getStyleRules(array $classes): array {
    $style_rules = [];

    $cached_items = $this->cache->getMultiple($classes);
    foreach ($cached_items as $class => $cached) {
      if (!empty($cached->data)) {
        $style_rules[] = $cached->data;
      }
    }

    if (empty($classes)) {
      return $style_rules;
    }

    $tailwind_styles = $this->ruleStorage->getRules($classes);

    // Log a warning if there are still missing classes
    // (ruleStorage->getRules() removes found items).
    if (\count($classes)) {
      $this->logger->warning('The following Tailwind classes could not be found %classes.', [
        '%classes' => \implode(', ', $classes),
      ]);
      // Cache negative classes.
      foreach ($classes as $class) {
        $this->cache->set($class, NULL);
      }
    }

    // Cache positive classes and add to output.
    foreach ($tailwind_styles as $class => $rules) {
      $this->cache->set($class, $rules);
      $style_rules[] = $rules;
    }

    return $style_rules;
  }

}
